///
/// Microsoft SQL Server 2005 Business Intelligence Metadata Reporting Samples
/// Dependency Analyzer Sample
/// 
/// Copyright (c) Microsoft Corporation.  All rights reserved.
///
using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Text;
using Microsoft.AnalysisServices;
using MetaDataRepository;
using TSQLParser;

namespace TestRepository
{
    /// <summary>
    ///  Enumerates Analysis Services objects and the relationships between them. 
    ///  The result of this enumeration is persisted in a repository.
    /// </summary>
    class SSASEnumerator
    {
        /// <summary>
        ///  For storing repository information
        /// </summary>
        private class DsvRepositoryInformation
        {
            public mdObject dsview;
            public Dictionary<string, mdObject> dsvTableNameToIDMap;
        }

        private MetaDataRepository.Repository repository;
        private mdExecution Execution;

        /// <summary>
        /// Types of Analysis Services objects we recognise for the repository
        /// </summary>
        private class ObjectTypes
        {
            internal const string Server = "Ssas.Analysis Server";
            internal const string Database = "Ssas.Database";
            internal const string Cube = "Ssas.Cube";
            internal const string CubeDimension = "Ssas.Cube Dimension";
            internal const string MeasureGroup = "Ssas.Measure Group";
            internal const string MeasureGroupDimension = "Ssas.Measure Group Dimension";
            internal const string Partition = "Ssas.Partition";
            internal const string DatabaseDimension = "Ssas.Database Dimension";
            internal const string DataSource = "Ssas.DataSource";
            internal const string DataSourceView = "Ssas.DSV";
            internal const string DataSourceViewTable = "Ssas.DSV Table";
            internal const string Table = "Table";
        }

        /// <summary>
        ///  since we can't get the quoting characters offline, we'll have to hard code them here.
        /// </summary>
        private class QuotingCharacters
        {
            public static string OpenQuote = "[";
            public static string CloseQuote = "]";
        }

        public bool Initialize(MetaDataRepository.Repository repository, mdExecution execution)
        {
            this.repository = repository;
            this.Execution = execution;

            //// add types of objects to the object types table
            //AddASObjectType(ObjectTypes.Server, "Analysis Server");
            //AddASObjectType(ObjectTypes.Database, "Database");
            //AddASObjectType(ObjectTypes.DataSource, "Data Source");
            //AddASObjectType(ObjectTypes.DataSourceView, "Data Source View");
            //AddASObjectType(ObjectTypes.DataSourceViewTable, "Data Source View Table");
            //AddASObjectType(ObjectTypes.Cube, "Cube");
            //AddASObjectType(ObjectTypes.MeasureGroup, "Measure Group");
            //AddASObjectType(ObjectTypes.Partition, "Partition");
            //AddASObjectType(ObjectTypes.DatabaseDimension, "Database Dimension");
            //AddASObjectType(ObjectTypes.CubeDimension, "Cube Dimension");
            //AddASObjectType(ObjectTypes.MeasureGroupDimension, "Measure Group Dimension");

            return true;
        }

        private String QuoteName(String objectName)
        {
            if (String.IsNullOrEmpty(objectName))
            {
                return String.Empty;
            }
            else
            {
                if (!objectName.Contains(QuotingCharacters.OpenQuote))
                {
                    return String.Format("{0}{1}{2}", QuotingCharacters.OpenQuote, objectName, QuotingCharacters.CloseQuote);
                }
                else
                {
                    return objectName;
                }
            }
        }

        //private void AddASObjectType(string type, string name)
        //{
        //    repository.AddObjectType(Repository.Domains.SSAS, type, name);
        //}
        
        /// <summary>
        ///  Opens a connection to the Analysis Server and does the actual object model walk
        /// </summary>
        /// <param name="connectionString"></param>
        public void EnumerateServer(string connectionString)
        {
            using (Microsoft.AnalysisServices.Server server = new Server())
            {
                server.Connect(connectionString);
                mdDatabaseServer dbServer = repository.GetDatabaseServer(Execution, server.Name, "msolap");
                mdDatabase database = null; 

                if (connectionString.ToLower().Contains("catalog"))
                {
                    Database db = server.Databases[server.ConnectionInfo.Catalog];
                    Console.Write(string.Format("Enumerating '{0}'...", db.Name));
                    database = repository.GetDatabase(Execution, dbServer, db.Name);
                    EnumerateDatabase(database, db);
                    Console.WriteLine("Done");
                }
                else
                {
                    // get the databases
                    foreach (Database db in server.Databases)
                    {
                        Console.Write(string.Format("Enumerating '{0}'...", db.Name));
                        database = repository.GetDatabase(Execution, dbServer, db.Name);
                        EnumerateDatabase(database, db);
                        Console.WriteLine("Done");
                    }
                }
            }

        }

        private void EnumerateDatabase(mdDatabase database, Database db)
        {
            Dictionary<DataSource, mdDatabase> dsToIdMap;

            EnumerateDataSources(db, database, out dsToIdMap);

            Dictionary<DataSourceView, DsvRepositoryInformation> dsvToObjectIDMap;
            EnumerateDataSourceViews(db, database, out dsvToObjectIDMap);

            Dictionary<Dimension, MetaDataRepository.mdObject> dimensionToIDMap;
            EnumerateDimensions(db, database, dsvToObjectIDMap, dsToIdMap, out dimensionToIDMap);

            EnumerateCubes(db, database, dsvToObjectIDMap, dsToIdMap, dimensionToIDMap);
        }

        #region EnumerateDataSources
        private void EnumerateDataSources(Database db, mdDatabase database, out Dictionary<DataSource, mdDatabase> dsToIdMap)
        {
            dsToIdMap = new Dictionary<DataSource, mdDatabase>(db.DataSources.Count);

            foreach (DataSource ds in db.DataSources)
            {

                //int dsID = repository.AddObject(ds.Name, ds.Description, ObjectTypes.DataSource, dbID);

                //// add connections   
                //int connectionID = repository.GetConnection(ds.ConnectionString);

                //if (connectionID == -1)
                //{
                //    connectionID = repository.AddObject(ds.Name, ds.Description, string.Empty, repository.RootRepositoryObjectID);
                //}

                dsToIdMap.Add(ds, repository.GetDatabase(Execution, ds.ConnectionString ));
            }
        }
        #endregion

        #region Data Source Views
        private void EnumerateDataSourceViews(Database db, mdDatabase database, out Dictionary<DataSourceView, DsvRepositoryInformation> dsvToObjectIDMap)
        {
            dsvToObjectIDMap = new Dictionary<DataSourceView, DsvRepositoryInformation>();

            foreach (DataSourceView dsv in db.DataSourceViews)
            {
                mdDatabase dsvDatabase = null;

                if (dsv.DataSource != null)
                {
                    dsvDatabase = repository.GetDatabase(Execution, dsv.DataSource.ConnectionString);
                }
                else
                {
                    dsvDatabase = database;
                }
                // a DSV might exist without a data source. So, we have to use the database as the parent.
                mdObject dsview = repository.GetObject(database, string.Empty, QuoteName(dsv.Name), ObjectTypes.DataSourceView);

                DsvRepositoryInformation dsvRepositoryInformation = new DsvRepositoryInformation();
                dsvRepositoryInformation.dsview = dsview;

                EnumerateDsvTables(dsv, database, dsvDatabase, dsview, out dsvRepositoryInformation.dsvTableNameToIDMap);

                dsvToObjectIDMap.Add(dsv, dsvRepositoryInformation);
            }
        }

        private void EnumerateDsvTables(DataSourceView dsv, mdDatabase database, mdDatabase dsvDatabase, mdObject dsView, out Dictionary<string, mdObject> tableNameToIDMap)
        {
            tableNameToIDMap = new Dictionary<string, mdObject>(dsv.Schema.Tables.Count); 

            foreach (DataTable table in dsv.Schema.Tables)
            {
                string tableType = GetExtendedProperty(table, ExtendedProperties.TableType);

                // If this doesn't have a type, then don't bother with it.
                if (!String.IsNullOrEmpty(tableType))
                {

                    // add the data source view's Table
                    mdObject dsvTable = repository.GetObject(database, QuoteName(GetSchemaName(table)), QuoteName(GetTableName(table)), tableType);
                    // Add the linkage to the dsView.
                    repository.GetObjectMap(dsView, dsvTable, true);
                    //int dsvTableID = repository.AddObject(table.TableName, string.Empty, ObjectTypes.DataSourceViewTable, dsvID);

                    tableNameToIDMap.Add(table.TableName, dsvTable);

                    if (dsvDatabase != null)
                    {
                        if (tableType == ExtendedProperties.TableTypeValue)
                        {
                            // this is mapped to an underlying table
                            mdObject dbTable = repository.GetObject(dsvDatabase, QuoteName(GetSchemaName(table)), QuoteName(GetTableName(table)), tableType);

                            // add the lineage
                            repository.GetObjectMap(dsvTable, dbTable, false);
                        }
                        else
                        {
                            if (tableType == ExtendedProperties.ViewTypeValue)
                            {
                                // add the query definition
                                object queryDefinition = table.ExtendedProperties["QueryDefinition"];
                                if (queryDefinition != null)
                                {
                                    ParseTSqlStatement(queryDefinition.ToString(), dsvDatabase, dsvTable);
                                    repository.UpdateObjectsQuery(dsvTable, queryDefinition.ToString());
                                }
                                else
                                {
                                    // And if there isn't a Query Definition, then check for a table name and use that instead.  id #14969
                                    if (!String.IsNullOrEmpty(table.TableName))
                                    {
                                        mdObject dbTable = repository.GetObject(dsvDatabase, QuoteName(GetSchemaName(table)), QuoteName(GetTableName(table)), tableType);

                                        // add the lineage
                                        repository.GetObjectMap(dsvTable, dbTable, false);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        #endregion

        #region Parse TSQL Statement
        private void ParseTSqlStatement(string statement, mdDatabase database, mdObject dsvTable)
        {
            SqlStatement toBeParsed = new SqlStatement();

            //toBeParsed.sqlString = statement;
            toBeParsed.quotedIdentifiers = true;
            if (toBeParsed.ParseString(statement))
            {
                foreach (string tableName in toBeParsed.getTableNames(true))
                {
                    String[] splitObject = tableName.Split('.');
                    mdObject dbTable = repository.GetObject(database, splitObject[splitObject.Length - 2], splitObject[splitObject.Length - 1], "Table");
                    repository.GetObjectMap(dsvTable, dbTable, false);
                }
                foreach (string procedureName in toBeParsed.getProcedureNames(true))
                {
                    String[] splitObject = procedureName.Split('.');
                    mdObject dbTable = repository.GetObject(database, splitObject[splitObject.Length - 2], splitObject[splitObject.Length - 1], "StoredProcedure");
                    repository.GetObjectMap(dsvTable, dbTable, false);
                }
                foreach (string funcName in toBeParsed.getFunctionNames(true))
                {
                    String[] splitObject = funcName.Split('.');
                    mdObject dbTable = repository.GetObject(database, splitObject[splitObject.Length - 2], splitObject[splitObject.Length - 1], "UserDefinedFunction");
                    repository.GetObjectMap(dsvTable, dbTable, false);
                }
            }
            else
            {
                string errorMessage = "The following messages where generated whilst parsing the sql statement\r\n" + statement + "\r\n";
                foreach (string error in toBeParsed.parseErrors)
                {
                    errorMessage += error + "\r\n";
                }
                Console.WriteLine(errorMessage);
            }
        }
        #endregion


        private void EnumerateDimensions(Database db, mdDatabase database, Dictionary<DataSourceView, DsvRepositoryInformation> dsvToObjectIDMap, Dictionary<DataSource, mdDatabase> dsToIDMap, out Dictionary<Dimension, MetaDataRepository.mdObject> dimensionToIdMap)
        {
            dimensionToIdMap = new Dictionary<Dimension, MetaDataRepository.mdObject>(db.Dimensions.Count);

            foreach (Dimension dim in db.Dimensions)
            {
                // get the connection ID
                mdDatabase connectionID = null;

                if ((dim.DataSource != null ) && (dsToIDMap.ContainsKey(dim.DataSource)))
                {
                    connectionID = dsToIDMap[dim.DataSource];
                }

                DsvRepositoryInformation dsvRepositoryInformation = dsvToObjectIDMap[dim.DataSourceView]; 
                Dictionary<string,MetaDataRepository.mdObject> dsvTableNameToIdMap = dsvRepositoryInformation.dsvTableNameToIDMap;

                mdObject dimObject = repository.GetObject(database, string.Empty, QuoteName(dim.Name), dim.Type.ToString(), dim.Description);
                //int dimID = repository.AddObject(dim.Name, dim.Description, ObjectTypes.DatabaseDimension, dbID);

                dimensionToIdMap.Add(dim, dimObject);

                Dictionary<MetaDataRepository.mdObject, bool> sourceTableIDs = new Dictionary<MetaDataRepository.mdObject, bool>();

                foreach (DimensionAttribute attr in dim.Attributes)
                {
                    if (connectionID != null)
                    {
                        string query; 
                        mdObject sourceTableID = GetSourceIDForBinding(connectionID, dsvTableNameToIdMap, attr.NameColumn.Source, out query);

                        if (sourceTableID != null)
                        {
                            // since we track only table level dependencies and not at the column level,
                            // we have to create a distinct list of these source tables
                            sourceTableIDs[sourceTableID] = true;
                        }
                        else
                        {
                            // if we don't have a table ID but have a query, store that as an attribute
                            // so that users can at least take a look at the query itself.
                            if (query != null)
                            {

                                // repository.AddAttribute(dimID, Repository.Attributes.QueryDefinition, query);
                                ParseTSqlStatement(query, connectionID, dimObject);
                                // Store the query against the view that it came from.
                                repository.UpdateObjectsQuery(dimObject, query);
                            }
                        }
                    }
                }

                // now create the lineage map from the distinct list of source tables to the dimension 
                foreach(mdObject sourceTableID in sourceTableIDs.Keys)
                {
                    //repository.AddMapping(sourceTableID, dimID);
                    repository.GetObjectMap(sourceTableID, dimObject, false);
                }
            }
        }

        /// <summary>
        ///  returns information about the binding for table/dsvTable or ColumnBinding, we return the table ID
        ///  for query binding, the returned value is -1, but the query out parameter is set to the query definition
        /// </summary>
        /// <param name="connectionID"></param>
        /// <param name="dsvTableNameToIdMap"></param>
        /// <param name="binding"></param>
        /// <param name="?"></param>
        /// <returns></returns>
        private mdObject GetSourceIDForBinding(mdDatabase connectionID, Dictionary<string, mdObject> dsvTableNameToIdMap, Binding binding, out string query)
        {
            mdObject sourceID = null;
            query = null;

            if (binding != null)
            {
                if (binding is ColumnBinding)
                {
                    ColumnBinding colBinding = (ColumnBinding)binding;

                    // get the id of the dsv table colBinding.TableID
                    if (dsvTableNameToIdMap.ContainsKey(colBinding.TableID))
                    {
                        sourceID = dsvTableNameToIdMap[colBinding.TableID];
                    }
                }
                else
                {
                    if (binding is DsvTableBinding)
                    {
                        DsvTableBinding dsvTableBinding = (DsvTableBinding)binding;

                        if (dsvTableNameToIdMap.ContainsKey(dsvTableBinding.TableID))
                        {
                            sourceID = dsvTableNameToIdMap[dsvTableBinding.TableID];
                        }
                    }
                    else
                    {
                        if (binding is TableBinding)
                        {
                            TableBinding tableBinding = (TableBinding)binding;

                            string tableName = GetFullyQualifiedTableName(tableBinding.DbSchemaName, tableBinding.DbTableName);

                            // get the table name from the repository itself
                            //sourceID = repository.GetTable(connectionID, tableName);
                            sourceID = repository.GetObject(connectionID, tableBinding.DbSchemaName, tableBinding.DbTableName, ObjectTypes.Table); // ToDo: Determine the correct type.

                            //if (sourceID == null)
                            //{
                            //    sourceID = repository.AddObject(tableName, string.Empty, RelationalEnumerator.ObjectTypes.Table, connectionID);
                            //}
                        }
                        else
                        {
                            if (binding is MeasureGroupDimensionBinding)
                            {
                                // the caller will handle this since we need the list of cb dimensions
                            }
                            else
                            {
                                if (binding is QueryBinding)
                                {
                                    QueryBinding queryBinding = (QueryBinding)binding;

                                    query = queryBinding.QueryDefinition;
                                }
                            }
                        }
                    }
                }
            }

            return sourceID;
        }

        private void EnumerateCubes(Database db, mdDatabase database, Dictionary<DataSourceView, DsvRepositoryInformation> dsvToRepositoryMap, Dictionary<DataSource, mdDatabase> dsToIdMap, Dictionary<Dimension, mdObject> dimensionToIDMap)
        {
            Dictionary<string, mdObject> emptyDsvTableToIdMap = new Dictionary<string, mdObject>();

            foreach (Cube cb in db.Cubes)
            {
                mdDatabase connectionID = dsToIdMap[cb.DataSource];

                // int cbID = repository.AddObject(cb.Name, cb.Description, ObjectTypes.Cube, dbID);
                mdObject cbObject = repository.GetObject(database, string.Empty, cb.Name, ObjectTypes.Cube, cb.Description);
                Dictionary<CubeDimension, mdObject> cubeDimensionToIDMap = new Dictionary<CubeDimension, mdObject>();

                foreach (CubeDimension cbDim in cb.Dimensions)
                {
                    mdObject cbDimID = repository.GetObject(database, string.Empty, cbDim.Name, ObjectTypes.CubeDimension, cbDim.Description); //AddObject(cbDim.Name, cbDim.Description, ObjectTypes.CubeDimension, cbID);

                    // add Use dependency between the cube dimension and the database dimension
                    //repository.AddUseDependency(dimensionToIDMap[cbDim.Dimension], cbDimID);
                    repository.GetObjectMap(cbObject, cbDimID, true);
                    repository.GetObjectMap(dimensionToIDMap[cbDim.Dimension], cbDimID, false);

                    cubeDimensionToIDMap.Add(cbDim, cbDimID);
                }

                EnumerateMeasureGroups(dsvToRepositoryMap, emptyDsvTableToIdMap, cb, connectionID, cbObject, cubeDimensionToIDMap);
            }
        }

        /// <summary>
        /// enumerates measures groups, mg dimensions and partitions.
        /// </summary>
        /// <param name="dsvToRepositoryMap"></param>
        /// <param name="emptyDsvTableToIdMap"></param>
        /// <param name="cb"></param>
        /// <param name="connectionID"></param>
        /// <param name="cbID"></param>
        /// <param name="cubeDimensionToIDMap"></param>
        private void EnumerateMeasureGroups(Dictionary<DataSourceView, DsvRepositoryInformation> dsvToRepositoryMap, Dictionary<string, mdObject> emptyDsvTableToIdMap, Cube cb, mdDatabase connectionID, mdObject cbID, Dictionary<CubeDimension, mdObject> cubeDimensionToIDMap)
        {
            // measure group
            foreach (MeasureGroup mg in cb.MeasureGroups)
            {
                //int mgID = repository.AddObject(mg.Name, mg.Description, ObjectTypes.MeasureGroup, cbID);
                mdObject mgObject = repository.GetObject(connectionID, string.Empty, QuoteName(mg.Name), ObjectTypes.MeasureGroup);

                // partition
                foreach (Partition pt in mg.Partitions)
                {
                    //int ptID = repository.AddObject(pt.Name, pt.Description, ObjectTypes.Partition, mgID);
                    mdObject ptObject = repository.GetObject(connectionID, string.Empty, QuoteName(pt.Name), ObjectTypes.Partition, pt.Description);
                    Dictionary<string, mdObject> dsvTableToIdMap = emptyDsvTableToIdMap;
                    repository.GetObjectMap(mgObject, ptObject, true);

                    if ((pt.DataSourceView != null) && (dsvToRepositoryMap.ContainsKey(pt.DataSourceView)))
                    {
                        dsvTableToIdMap = dsvToRepositoryMap[pt.DataSourceView].dsvTableNameToIDMap;
                    }

                    // add the mapping

                    string query;
                    mdObject sourceTableID = GetSourceIDForBinding(connectionID, dsvTableToIdMap, pt.Source, out query);

                    if (sourceTableID != null)
                    {
                        repository.GetObjectMap(sourceTableID, ptObject, false);
                        //repository.AddMapping(sourceTableID, ptID);
                    }
                    else
                    {
                        // if we don't have a table ID but have a query, store that as an attribute
                        // so that users can at least take a look at the query itself.

                        if (query != null)
                        {
                            ParseTSqlStatement(query, connectionID, ptObject);
                            //repository.AddAttribute(ptID, Repository.Attributes.QueryDefinition, query);
                            repository.UpdateObjectsQuery(ptObject, query);
                        }
                    }
                }

                EnumerateMeasureGroupDimensions(dsvToRepositoryMap, connectionID, cubeDimensionToIDMap, mg, mgObject);
            }
        }

        private void EnumerateMeasureGroupDimensions(Dictionary<DataSourceView, DsvRepositoryInformation> dsvToRepositoryMap, mdDatabase connectionID, Dictionary<CubeDimension, mdObject> cubeDimensionToIDMap, MeasureGroup mg, mdObject mgID)
        {
            foreach (MeasureGroupDimension mgDim in mg.Dimensions)
            {
                if (mgDim.Dimension != null)
                {
                    Dictionary<string, mdObject> dsvTableNameToIdMap = dsvToRepositoryMap[mgDim.Dimension.DataSourceView].dsvTableNameToIDMap;
                    Dictionary<mdObject, bool> distinctTableIDsList = new Dictionary<mdObject, bool>();

                    mdObject mgDimID = repository.GetObject(connectionID, string.Empty, QuoteName(mgDim.Dimension.Name), ObjectTypes.MeasureGroupDimension, mgDim.Dimension.Description); 
                    // .AddObject(mgDim.Dimension.Name, mgDim.Dimension.Description, ObjectTypes.MeasureGroupDimension, mgID);

                    // add a Use dependency from cubeDim->mgdim
                    //repository.AddUseDependency(cubeDimensionToIDMap[mgDim.CubeDimension], mgDimID);
                    repository.GetObjectMap(cubeDimensionToIDMap[mgDim.CubeDimension], mgDimID, false);

                    // this handles ReferenceMeasureGroupDimension and DegenerateMeasureGroupDimension as well.
                    RegularMeasureGroupDimension rmgdim = mgDim as RegularMeasureGroupDimension;

                    if (rmgdim != null)
                    {
                        foreach (MeasureGroupAttribute mgDimAttr in rmgdim.Attributes)
                        {
                            foreach (DataItem keyColumn in mgDimAttr.KeyColumns)
                            {
                                string query;

                                mdObject sourceTableID = GetSourceIDForBinding(connectionID, dsvTableNameToIdMap, keyColumn.Source, out query);

                                if (sourceTableID != null)
                                {
                                    distinctTableIDsList[sourceTableID] = true;
                                }
                                else
                                {
                                    // if we don't have a table ID but have a query, store that as an attribute
                                    // so that users can at least take a look at the query itself.
                                    if (query != null)
                                    {
                                        ParseTSqlStatement(query, connectionID, mgDimID);
                                        //repository.AddAttribute(mgDimID, Repository.Attributes.QueryDefinition, query);
                                        repository.UpdateObjectsQuery(mgDimID, query);
                                    }
                                }
                            }
                        }
                    }

                    // add a lineage map from the tables to the measure group dim
                    foreach (mdObject tableID in distinctTableIDsList.Keys)
                    {
                        //repository.AddMapping(tableID, mgDimID);
                        repository.GetObjectMap(tableID, mgDimID, false);
                    }
                }
            }
        }


        private string GetFullyQualifiedTableName(DataTable dataTable)
        {
            return GetFullyQualifiedTableName(GetSchemaName(dataTable), GetTableName(dataTable));
        }

        private string GetFullyQualifiedTableName(string schemaName, string tableName)
        {
            return EncodeTableFullName(schemaName, tableName, QuotingCharacters.OpenQuote, QuotingCharacters.CloseQuote);
        }
        
        /// <summary>
        /// Helper function encode full table name
        /// </summary>
        /// <param name="schemaName"></param>
        /// <param name="objName"></param>
        /// <param name="openQuote"></param>
        /// <param name="closeQuote"></param>
        /// <returns></returns>
        private string EncodeTableFullName(string schemaName, string objName, string openQuote, string closeQuote)
        {
            if (schemaName == null || schemaName.Length == 0)
                return EncodeObjectName(objName, openQuote, closeQuote);
            else
                return (String.Format("{0}.{1}", EncodeObjectName(schemaName, openQuote, closeQuote),
                    EncodeObjectName(objName, openQuote, closeQuote)));
        }

        /// <summary>
        /// Encode Object Name
        /// </summary>
        /// <param name="objectName">Object Name</param>
        /// <returns>Encoded Object Name</returns>
        public static string EncodeObjectName(string objectName, string openQuote, string closeQuote)
        {
            Debug.Assert(openQuote != null, "Undefined open quote.");
            Debug.Assert(closeQuote != null, "Undefined close quote.");

            if (objectName == null)
                return "";

            if (objectName.Trim().Length == 0)
                return "";

            if (openQuote == null || closeQuote == null)
                return objectName;

            string newName = objectName.Replace(closeQuote, closeQuote + closeQuote);
            return (String.Format("{0}{1}{2}", openQuote, newName, closeQuote));
        }

        /// <summary>
        /// Gets the Schema name for the given dataTable.
        /// </summary>
        /// <param name="dataTable">Data Table</param>
        /// <returns>Schema Name</returns>
        private string GetSchemaName(DataTable dataTable)
        {
            return GetExtendedProperty(dataTable, ExtendedProperties.DBSchemaName);
        }

        /// <summary>
        /// Gets the Table name for the given dataTable.
        /// </summary>
        /// <param name="dataTable">Data Table</param>
        /// <returns>Table Name</returns>
        private string GetTableName(DataTable dataTable)
        {

            string tableName = GetExtendedProperty(dataTable, ExtendedProperties.DBTableName);
            if (tableName == null || tableName.Length == 0)
                return dataTable.TableName;

            return tableName;

        }

        /// <summary>
        /// Get Extended Property from Object
        /// </summary>
        /// <param name="obj">Data Object</param>
        /// <param name="propName">Property Name</param>
        /// <returns>Property Value</returns>
        public static string GetExtendedProperty(object obj, string propName)
        {
            if (obj is System.Data.DataSet)
            {
                System.Data.DataSet dataSet = obj as System.Data.DataSet;
                if (dataSet.ExtendedProperties.Contains(propName))
                    return (string)dataSet.ExtendedProperties[propName];

                return null;
            }

            if (obj is System.Data.DataTable)
            {
                System.Data.DataTable dataTable = obj as System.Data.DataTable;
                if (dataTable.ExtendedProperties.Contains(propName))
                    return (string)dataTable.ExtendedProperties[propName];

                return null;
            }

            if (obj is System.Data.DataColumn)
            {
                System.Data.DataColumn dataColumn = obj as System.Data.DataColumn;
                if (dataColumn.ExtendedProperties.Contains(propName))
                    return (string)dataColumn.ExtendedProperties[propName];

                return null;
            }

            if (obj is System.Data.DataRelation)
            {
                System.Data.DataRelation dataRelation = obj as System.Data.DataRelation;
                if (dataRelation.ExtendedProperties.Contains(propName))
                    return (string)dataRelation.ExtendedProperties[propName];

                return null;
            }

            if (obj is System.Data.UniqueConstraint)
            {
                System.Data.UniqueConstraint uc = obj as System.Data.UniqueConstraint;
                if (uc.ExtendedProperties.Contains(propName))
                    return (string)uc.ExtendedProperties[propName];

                return null;
            }

            if (obj is System.Data.ForeignKeyConstraint)
            {
                System.Data.ForeignKeyConstraint fkc = obj as System.Data.ForeignKeyConstraint;
                if (fkc.ExtendedProperties.Contains(propName))
                    return (string)fkc.ExtendedProperties[propName];

                return null;
            }
            return null;

        }
    
    }

    /// <summary>
    /// Serves as an enumeration of Extended property names 
    /// </summary>
    public class ExtendedProperties
    {
        private ExtendedProperties() { }

        public const string DataSourceID = "DataSourceID";
        public const string DBTableName = "DbTableName";
        public const string DBSchemaName = "DbSchemaName";
        public const string DBColumnName = "DbColumnName";
        public const string FriendlyName = "FriendlyName";
        public const string Description = "Description";
        public const string BatchID = "BatchID";
        public const string IsError = "IsError";
        public const string IsLogical = "IsLogical";
        public const string TableType = "TableType";
        public const string QueryDefinition = "QueryDefinition";
        public const string QueryBuilder = "QueryBuilder";
        public const string GenericQueryBuilderValue = "GenericQueryBuilder";
        public const string SpecificQueryBuilderValue = "SpecificQueryBuilder";
        public const string ComputedColumnExpression = "ComputedColumnExpression";
        public const string DataSize = "DataSize";
        public const string NewName = "NewName";
        public const string Regenerating = "Regenerating";
        public const string AllowGen = "AllowGen";
        public const string Project = "Project";
        public const string DSVTable = "DSVTable";
        public const string DSVColumn = "DSVColumn";
        public const string DSVRelation = "DSVRelation";
        public const string TrueValue = "True";
        public const string FalseValue = "False";
        public const string ViewTypeValue = "View";
        public const string TableTypeValue = "Table";
    }		

}
